#ifndef _TEST_TOOL_HPP_
#define _TEST_TOOL_HPP_
/**
 * @file test_tool.hpp
 *
 * @brief some tools for CUDA
 */
#include <cuda.h>

inline void exception_maker(cudaError_t rc, const char * funcname)
{
    using namespace std;
    if (rc != cudaSuccess) {
	const char * message = cudaGetErrorString(rc);
	fprintf(stderr, "In %s Error:%s\n", funcname, message);
	throw message;
    }
}

inline int ccudaMalloc(void **devPtr, size_t size)
{
    cudaError_t rc = cudaMalloc((void **)(void*)devPtr, size);
    exception_maker(rc, "ccudaMalloc");
    return cudaSuccess;
}

inline int ccudaFree(void *devPtr)
{
    cudaError_t rc = cudaFree(devPtr);
    exception_maker(rc, "ccudaFree");
    return cudaSuccess;
}

inline int ccudaMemcpy(void *dest, void *src, size_t size,
		      enum cudaMemcpyKind kind)
{
    cudaError_t rc = cudaMemcpy(dest, src, size, kind);
    exception_maker(rc, "ccudaMemcpy");
    return cudaSuccess;
}

inline int ccudaEventCreate(cudaEvent_t * event)
{
    cudaError_t rc = cudaEventCreate(event);
    exception_maker(rc, "ccudaEventCreate");
    return cudaSuccess;
}

inline int ccudaEventRecord(cudaEvent_t event, cudaStream_t stream)
{
    cudaError_t rc = cudaEventRecord(event, stream);
    exception_maker(rc, "ccudaEventRecord");
    return cudaSuccess;
}

inline int ccudaEventSynchronize(cudaEvent_t event)
{
    cudaError_t rc = cudaEventSynchronize(event);
    exception_maker(rc, "ccudaEventSynchronize");
    return cudaSuccess;
}

inline int ccudaThreadSynchronize()
{
    cudaError_t rc = cudaThreadSynchronize();
    exception_maker(rc, "ccudaThreadSynchronize");
    return cudaSuccess;
}

inline int ccudaEventElapsedTime(float * ms,
				 cudaEvent_t start, cudaEvent_t end)
{
    cudaError_t rc = cudaEventElapsedTime(ms, start, end);
    exception_maker(rc, "ccudaEventElapsedTime");
    return cudaSuccess;
}

inline int ccudaEventDestroy(cudaEvent_t event)
{
    cudaError_t rc = cudaEventDestroy(event);
    exception_maker(rc, "ccudaEventDestroy");
    return cudaSuccess;
}

inline int ccudaMemcpyToSymbol(const void * symbol,
			       const void * src,
			       size_t count,
			       size_t offset = 0,
			       enum cudaMemcpyKind kind
			       = cudaMemcpyHostToDevice)
{
    cudaError_t rc = cudaMemcpyToSymbol((const char *)symbol,
					src, count, offset, kind);
    exception_maker(rc, "ccudaMemcpyToSymbol");
    return cudaSuccess;
}

template<class T, int dim, enum cudaTextureReadMode readMode>
inline int ccudaBindTexture(size_t * offset,
			    const struct texture< T, dim, readMode > & texref,
			    const void * devPtr,
			    size_t size = UINT_MAX)
{

    cudaError_t rc = cudaBindTexture(offset, texref, devPtr, size);
    exception_maker(rc, "ccudaBIndTexture");
    return cudaSuccess;
}

inline int ccudaSetDevice(int dev)
{
    cudaError rc = cudaSetDevice(dev);
    exception_maker(rc, "ccudaSetDevice");
    return cudaSuccess;
}

/**
 * This function is used to compare the outputs with C program's.
 * @param array data to be printed.
 * @param size size of array.
 * @param threads number of threads.
 */
static inline void print_float_array(const float array[], int size, int threads)
{
    int b = size / threads;
    int i;
    int j;

    /* Comment for not English native like me,
       'showpoint' means show zeros after point
       like 1.00000.
    */
    //cout << setw(12) << dec << setprecision(10) << showpoint;
    //cout << showpoint;
    //cout << dec;
    //cout << setprecision(10);
    //After wasting huge time, I gave up using C++ style.
    for (j = 0; j < 5; j += 5) {
	//cout << setw(12) << array[j + 0] << " ";
	printf("%.10f ", array[j + 0]);
	printf("%.10f ", array[j + 1]);
	printf("%.10f ", array[j + 2]);
	printf("%.10f ", array[j + 3]);
	printf("%.10f ", array[j + 4]);
	printf("\n");
    }
    for (i = 1; i < threads; i++) {
	for (j = -5; j < 5; j += 5) {
	    printf("%.10f ", array[b * i + j + 0]);
	    printf("%.10f ", array[b * i + j + 1]);
	    printf("%.10f ", array[b * i + j + 2]);
	    printf("%.10f ", array[b * i + j + 3]);
	    printf("%.10f ", array[b * i + j + 4]);
	    printf("\n");
	}
    }
    for (j = -5; j < 0; j += 5) {
	printf("%.10f ", array[size + j + 0]);
	printf("%.10f ", array[size + j + 1]);
	printf("%.10f ", array[size + j + 2]);
	printf("%.10f ", array[size + j + 3]);
	printf("%.10f ", array[size + j + 4]);
	printf("\n");
    }
}

/**
 * This function is used to compare the outputs with C program's.
 * @param array data to be printed.
 * @param size size of array.
 * @param threads number of threads.
 */
static inline void print_double_array(const double array[],
				      int size, int threads)
{
    int b = size / threads;
    int i;
    int j;

    for (j = 0; j < 4; j += 4) {
	printf("%.15f ", array[j + 0]);
	printf("%.15f ", array[j + 1]);
	printf("%.15f ", array[j + 2]);
	printf("%.15f ", array[j + 3]);
	printf("\n");
    }
    for (i = 1; i < threads; i++) {
	for (j = -4; j < 4; j += 4) {
	    printf("%.15f ", array[b * i + j + 0]);
	    printf("%.15f ", array[b * i + j + 1]);
	    printf("%.15f ", array[b * i + j + 2]);
	    printf("%.15f ", array[b * i + j + 3]);
	    printf("\n");
	}
    }
    for (j = -4; j < 0; j += 4) {
	printf("%.15f ", array[size + j + 0]);
	printf("%.15f ", array[size + j + 1]);
	printf("%.15f ", array[size + j + 2]);
	printf("%.15f ", array[size + j + 3]);
	printf("\n");
    }
}

#endif
